@@ -1,21 +1,45 @@
 from ..utils import utils
-import numpy as np
-from pyswarms.single.global_best import GlobalBestPSO
 from .csip_access import ServiceClient
+from .csip_access import csip_worker
+from pyswarms.single.global_best import GlobalBestPSO
 from os import path
+from threading import Thread
+import numpy as np
 import copy
-# import json
 import datetime
+import queue
+
+req_queue = queue.Queue()
+
+def service1(x, param, url, step_param_names, step_objfunc, calib_params, full_trace, files=None):
+    cost = np.ones(len(x[:, 0]))
+    res_queue = queue.Queue()
+
+    print('   ', end='', flush=True)
+
+    # submit for processing
+    for i_particle, v in enumerate(x[:, 0]):
+        req_queue.put((i_particle, x, step_param_names, calib_params, step_objfunc, res_queue))
+
+    # wait for the cost value to come back
+    for i, v in enumerate(x[:, 0]):
+        (i_particle, p_cost) = res_queue.get()
+        cost[i_particle] = p_cost
+        res_queue.task_done()
+
+    res_queue.join()
+    print(flush=True)
+    return cost
 
 
-def service(x, param, url, step_param_names, step_objfunc, calib_params, files=None):
+def service(x, param, url, step_param_names, step_objfunc, calib_params, full_trace, files=None):
     cost = np.ones(len(x[:, 0]))
     threads = []
     print('   ', end='', flush=True)
 
     for i_particle, v in enumerate(x[:, 0]):
         worker = ServiceClient(url, i_particle, cost, x, param,
-                               step_param_names, step_objfunc, calib_params, files)
+                               step_param_names, step_objfunc, calib_params, files, full_trace)
         threads.append(worker)
         worker.start()
         # stagger the calls a bit (250ms)
@@ -28,10 +52,9 @@ def service(x, param, url, step_param_names, step_objfunc, calib_params, files=N
     return cost
 
 
-def global_best(steps, rounds, args, n_particles, iters, options, rtol=1.0E-6, ftol=-np.inf):
-    """Performs stepwise PSO.
-
-        Performs a stepwise particle swarm optimization
+def global_best(steps, rounds, args, n_particles, iters, options, n_threads=4, rtol=0.001, ftol=-np.inf,
+                full_trace=None, old=True):
+    """Performs a stepwise particle swarm optimization
 
         Parameters
         ----------
@@ -48,11 +71,14 @@ def global_best(steps, rounds, args, n_particles, iters, options, rtol=1.0E-6, f
         options : dict
             PSO options (see pyswarms)
         rtol : float
-            relative error in objective_func(best_pos) acceptable for
-            convergence. Default is :code:`-np.inf`
+            percentage of change of sum(best_cost) between rounds for
+            convergence. Default is :code:`0.001` (0.1 percent)
         ftol : float
             PSO tolerance
-
+        full_trace : Array
+            trace of all runs list of tuples
+            first is dictionary of parameter names to parameter values
+            second is the cost value
         Returns
         -------
         tuple
@@ -79,7 +105,18 @@ def global_best(steps, rounds, args, n_particles, iters, options, rtol=1.0E-6, f
 
     # trace of steps info
     step_trace = {}
-    start_time =  datetime.datetime.now()
+    # best round cost
+    best_round_cost = 1000
+
+    done = False
+    workers = []
+    for thread_no in range(n_threads):
+        w = Thread(target=csip_worker, args=(req_queue, thread_no, lambda: done,
+                                             full_trace, args['url'], args['files'], args['param']))
+        workers.append(w)
+        w.start()
+
+    start_time = datetime.datetime.now()
     for r in range(max_rounds):
         no_improvement = np.full(len(steps), True)
         for s, step in enumerate(steps):
@@ -95,44 +132,66 @@ def global_best(steps, rounds, args, n_particles, iters, options, rtol=1.0E-6, f
             args['step_objfunc'] = objfunc
             # get calibrated parameter from all other steps
             args['calib_params'] = utils.get_calibrated_params(steps, s)
+            args['full_trace'] = full_trace
 
             # create optimizer in the first round.
             if optimizer[s] is None:
                 optimizer[s] = GlobalBestPSO(n_particles, len(param_names),
                                              options=options, bounds=bounds, ftol=ftol)
             print('\n>>>>> R{}/S{}  particle params: {}  calibrated params: {}\n'.format(r + 1, s + 1, param_names,
-                                                                             args['calib_params']))
+                                                                                         args['calib_params']))
 
             # perform optimization
-            cost, pos = optimizer[s].optimize(service, iters=iters, **args)
+            if old:
+                cost, pos = optimizer[s].optimize(service, iters=iters, **args)
+            else:
+                cost, pos = optimizer[s].optimize(service1, iters=iters, **args)
+
 
             # capture the best cost
-            if cost < best_cost[s] and np.abs(cost - best_cost[s]) > rtol:
+            # if cost < best_cost[s] and np.abs(cost - best_cost[s]) > rtol:
+            if cost < best_cost[s]:
                 best_cost[s] = cost
                 no_improvement[s] = False
                 utils.annotate_step(best_cost[s], pos, steps, s)
 
             print('\n     Step summary, best particle values: {} '.format(pos))
 
-            key = "r{}s{}".format(r+1, s+1)
-            step_trace[key]= copy.deepcopy(steps)
+            key = "r{}s{}".format(r + 1, s + 1)
+            step_trace[key] = copy.deepcopy(steps)
 
             # print(json.dumps(steps, sort_keys=False, indent=2))
 
         round_cost = np.sum(best_cost)
         # if no improvement in all steps, break out of rounds prematurely
-        print('\n  Round summary, round_cost:{}, step_costs: {}, improvement:{} '.format(round_cost, best_cost, np.invert(no_improvement)))
         # but start checking only after min_rounds
-        if (r + 1 >= min_rounds) and all(no_improvement):
+        # if (r + 1 >= min_rounds) and all(no_improvement):
+        rel_round_tol = 1 - round_cost / best_round_cost
+
+        print('\n  Round summary - round_cost:{}, step_costs: {}, step improvement:{}'
+              .format(round_cost, best_cost, np.invert(no_improvement)))
+        print('\n  Progress -  best_round_cost:{}, rel_round_tol:{}, rtol:{}'
+              .format(best_round_cost, rel_round_tol, rtol))
+
+        if (r + 1 >= min_rounds) and 0 <= rel_round_tol < rtol:
             break
 
-    end_time = datetime.datetime.now()
-    print('Done in {} after {} out of {} rounds'.format(str(end_time - start_time), r + 1, max_rounds))
+        if round_cost < best_round_cost:
+            best_round_cost = round_cost
 
-    step_trace['rounds'] = r+1
+    end_time = datetime.datetime.now()
+    elapsed = str(end_time - start_time)
+
+    print('Done in {} after {} out of {} rounds'.format(elapsed, r + 1, max_rounds))
+
+    done = True
+    for w in workers:
+        w.join()
+
+    step_trace['rounds'] = r + 1
     step_trace['steps'] = len(steps)
     step_trace['iters'] = iters
     step_trace['particles'] = n_particles
-    step_trace['time'] = str(end_time - start_time)
+    step_trace['time'] = elapsed
 
-    return optimizer, cost, step_trace  # return the last output
+    return optimizer, step_trace
