@@ -12,6 +12,14 @@ NORM = {
     'pbias': lambda x: abs(x)
 }
 
+def get_cost(response, objfunc):
+    """aggregated objective function value -> cost"""
+    cost = 0.0
+    for of in objfunc:
+        of_val = response.get_data_value(of['name'])
+        cost += NORM[of['name']](of_val) * of.get('weight', 1.0)
+    return cost
+
 
 def save(file, dict):
     """ Save dict to json file. """
